@@ -25,7 +25,7 @@
 ///        In-depth description of this algorithm:
 ///        https://github.com/kimwalisch/primecount/blob/master/doc/Easy-Special-Leaves.md
 ///
-/// Copyright (C) 2023 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2024 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -48,7 +48,6 @@
 
 #include <stdint.h>
 
-using std::numeric_limits;
 using namespace primecount;
 
 namespace {
@@ -359,13 +358,6 @@ T AC_OpenMP(T x,
   //
   #pragma omp parallel num_threads(threads) reduction(+: sum)
   {
-    // SegmentedPiTable is accessed very frequently.
-    // In order to get good performance it is important that
-    // SegmentedPiTable fits into the CPU's cache.
-    // Hence we use a small segment_size of x^(1/4).
-    SegmentedPiTable segmentedPi;
-    int64_t low, high;
-
     // C1 formula: pi[(x/z)^(1/3)] < b <= pi[pi_sqrtz]
     // There are very few iterations in this loop,
     // hence the use of an atomic loop counter (min_c1)
@@ -381,53 +373,78 @@ T AC_OpenMP(T x,
       sum -= C1<-1>(xp, b, b, pi_y, 1, min_m, max_m, primes, pi);
     }
 
-    // for (low = 0; low < sqrt; low += segment_size)
-    while (loadBalancer.get_work(low, high))
+    // SegmentedPiTable is accessed very frequently.
+    // In order to get good performance it is important that
+    // SegmentedPiTable fits into the CPU's cache.
+    // Hence we use a small segment_size of x^(1/4).
+    SegmentedPiTable segmentedPi;
+    ThreadDataAC thread;
+
+    // for (low = 0; low < sqrt(x); low += segment_size)
+    while (loadBalancer.get_work(thread))
     {
-      // Current segment [low, high[
-      segmentedPi.init(low, high);
-      T xlow = x / max(low, 1);
-      T xhigh = x / high;
+      int64_t low = thread.low;
+      int64_t segment_size = thread.segment_size;
+      int64_t limit = low + thread.segments * segment_size;
+      limit = std::min(limit, sqrtx);
 
-      int64_t min_c2 = max(k, pi_root3_xy);
-      min_c2 = max(min_c2, pi_sqrtz);
-      min_c2 = max(min_c2, pi[isqrt(low)]);
-      min_c2 = max(min_c2, pi[min(xhigh / y, x_star)]);
-      min_c2 += 1;
-
-      int64_t min_a = min(xhigh / high, x13);
-      min_a = pi[max(x_star, min_a)] + 1;
-
-      // Upper bound of A & C2 formulas:
-      // x / (p * q) >= low
-      // p * next_prime(p) <= x / low
-      // p <= sqrt(x / low)
-      T sqrt_xlow = isqrt(xlow);
-      int64_t max_c2 = pi[min(sqrt_xlow, x_star)];
-      int64_t max_a = pi[min(sqrt_xlow, x13)];
-
-      // C2 formula: pi[sqrt(z)] < b <= pi[x_star]
-      for (int64_t b = min_c2; b <= max_c2; b++)
+      for (; low < limit; low += segment_size)
       {
-        int64_t prime = primes[b];
-        T xp = x / prime;
+        // Current segment [low, high[
+        int64_t high = low + segment_size;
+        high = std::min(high, sqrtx);
+        segmentedPi.init(low, high);
 
-        if (xp <= numeric_limits<uint64_t>::max())
-          sum += C2_64(xlow, xhigh, (uint64_t) xp, y, b, prime, lprimes, pi, segmentedPi);
-        else
-          sum += C2_128(xlow, xhigh, xp, y, b, primes, pi, segmentedPi);
-      }
+        // We measure the thread computation time excluding the
+        // first expensive initialization of the segmentedPi
+        // lookup table. If the thread computation time is close
+        // to 0 then we increase the number of segments in the
+        // loadBalancer which should improve performance.
+        if (low == thread.low)
+          thread.secs = get_time();
 
-      // A formula: pi[x_star] < b <= pi[x13]
-      for (int64_t b = min_a; b <= max_a; b++)
-      {
-        int64_t prime = primes[b];
-        T xp = x / prime;
+        T xlow = x / max(low, 1);
+        T xhigh = x / high;
+        int64_t min_c2 = max(k, pi_root3_xy);
+        min_c2 = max(min_c2, pi_sqrtz);
+        min_c2 = max(min_c2, pi[isqrt(low)]);
+        min_c2 = max(min_c2, pi[min(xhigh / y, x_star)]);
+        min_c2 += 1;
 
-        if (xp <= numeric_limits<uint64_t>::max())
-          sum += A_64(xlow, xhigh, (uint64_t) xp, y, prime, lprimes, pi, segmentedPi);
-        else
-          sum += A_128(xlow, xhigh, xp, y, prime, primes, pi, segmentedPi);
+        int64_t min_a = min(xhigh / high, x13);
+        min_a = pi[max(x_star, min_a)] + 1;
+
+        // Upper bound of A & C2 formulas:
+        // x / (p * q) >= low
+        // p * next_prime(p) <= x / low
+        // p <= sqrt(x / low)
+        T sqrt_xlow = isqrt(xlow);
+        int64_t max_c2 = pi[min(sqrt_xlow, x_star)];
+        int64_t max_a = pi[min(sqrt_xlow, x13)];
+
+        // C2 formula: pi[sqrt(z)] < b <= pi[x_star]
+        for (int64_t b = min_c2; b <= max_c2; b++)
+        {
+          int64_t prime = primes[b];
+          T xp = x / prime;
+
+          if (xp <= port::numeric_limits<uint64_t>::max())
+            sum += C2_64(xlow, xhigh, (uint64_t) xp, y, b, prime, lprimes, pi, segmentedPi);
+          else
+            sum += C2_128(xlow, xhigh, xp, y, b, primes, pi, segmentedPi);
+        }
+
+        // A formula: pi[x_star] < b <= pi[x13]
+        for (int64_t b = min_a; b <= max_a; b++)
+        {
+          int64_t prime = primes[b];
+          T xp = x / prime;
+
+          if (xp <= port::numeric_limits<uint64_t>::max())
+            sum += A_64(xlow, xhigh, (uint64_t) xp, y, prime, lprimes, pi, segmentedPi);
+          else
+            sum += A_128(xlow, xhigh, xp, y, prime, primes, pi, segmentedPi);
+        }
       }
     }
   }
@@ -496,7 +513,7 @@ int128_t AC(int128_t x,
   int128_t sum;
 
   // uses less memory
-  if (max_prime <= numeric_limits<uint32_t>::max())
+  if (max_prime <= port::numeric_limits<uint32_t>::max())
   {
     auto primes = generate_primes<uint32_t>(max_prime);
     sum = AC_OpenMP((uint128_t) x, y, z, k, x_star, max_a_prime, primes, threads, is_print);
