@@ -3,7 +3,7 @@
 ///        This file contains helper functions and global variables
 ///        that are initialized with default settings.
 ///
-/// Copyright (C) 2022 Kim Walisch, <kim.walisch@gmail.com>
+/// Copyright (C) 2024 Kim Walisch, <kim.walisch@gmail.com>
 ///
 /// This file is distributed under the BSD License. See the COPYING
 /// file in the top level directory.
@@ -15,12 +15,10 @@
 #include <int128_t.hpp>
 #include <imath.hpp>
 #include <macros.hpp>
-#include <to_string.hpp>
 
 #include <algorithm>
 #include <chrono>
 #include <cmath>
-#include <limits>
 #include <string>
 #include <stdint.h>
 #include <utility>
@@ -61,7 +59,7 @@ namespace primecount {
 /// standard int128_t type is missing in <stdint.h>. We need to
 /// define a few functions that are not supported by the C++ STL.
 ///
-#if defined(HAVE_NON_STANDARD__INT128_T)
+#if defined(ENABLE_INT128_TO_STRING)
 
 std::string to_string(uint128_t n)
 {
@@ -115,7 +113,7 @@ maxint_t to_maxint(const std::string& expr)
     if (pos != std::string::npos)
     {
       std::string n = expr.substr(pos);
-      maxint_t limit = std::numeric_limits<maxint_t>::max();
+      maxint_t limit = port::numeric_limits<maxint_t>::max();
       std::string max_n = to_string(limit);
 
       if (n.size() > max_n.size() ||
